'use strict';

/* Dependencies */
const debug   = require('debug')('actjs')
  , assert    = require('assert')
  , boot      = require('./boot')
  , compose   = require('./compose')
  , findup    = require('findup-sync')
  , path      = require('path')
  , sort      = require('./sort')
  , Container = require('./container');

/**
 * @class Act
 */
class Act {
  /**
   * The class constructor sets default values for our application.
   * These are used throughout the system.
   * @method  constructor
   */
  constructor() {
    /**
     * The environment is used to load specific configuration
     * for plugs. These configuration files are used to modify
     * plugs in an easy way.
     * @type {String}
     */
    this.env = process.env.NODE_ENV || 'development';

    /**
     * Set the rootPath used throughout the entire application.
     * @type {String}
     */
    this.cwd = process.env.ACT_CWD || process.cwd();

    /**
     * Create a new Container object which resolves dependencies for
     * services and wires. Also sets the current Act instance so it
     * can be used in DI
     * @type {Container}
     */
    this.container = new Container().set('act', this);

    /* @type {Array} */
    this.wires = [];

    /* Debug information */
    debug('Environment: %s', this.env);
    debug('Current working directory: %s', this.cwd);
  }

  /**
   * Return an object of all default paths for an ActJS project.
   * @return {Object}
   */
  get paths() {
    return {
      client: path.join(this.cwd, 'client'),
      common: path.join(this.cwd, 'common'),
      server: path.join(this.cwd, 'server'),
    };
  }

  /**
   * @method boot
   * @param  {Array} include
   * @param  {Array} exclude
   * @return {Act}
   */
  boot(include, exclude) {
    debug('Booting application');

    /* @type {String} */
    let fileName = findup('package.json', {cwd: this.cwd});
    debug('Loading %s', fileName);

    /* @type {Object} */
    const options = {
      cwd:     this.cwd,
      types:   this.env !== 'production' ? ['dependencies', 'devDependencies'] : ['dependencies'],
      include: include,
      exclude: exclude,
    };

    /* Loop middlewares and add them */
    boot(fileName, options).forEach(wire => this.use(createInstance(wire)));

    /* Return fluent interface */
    return this;
  }

  /**
   * [use description]
   * @param  {Object[]} wire
   * @return {Act}
   */
  use(wire) {
    /* Also accept array */
    if (Array.isArray(wire)) {
      wire.forEach(wire => this.use(wire));
      return this;
    }

    /* Assert input */
    assert(wire && (typeof wire === 'function' || typeof wire === 'object'), `Invalid type specified: ${typeof wire}`);
    assert(typeof wire.code === 'string', `wire.code should be a valid string, ${typeof wire.code} given`);

    /* Set the wire in our context */
    this.container.set(wire.code, wire);

    /* Collect default resolvers */
    this
      .define(wire, wire.services, false)
      .define(wire, wire.constants, true);

    /* Push the middleware to the stack */
    this.wires.push(wire.code);

    debug('Added `%s` to the stack', wire.code);
    return this;
  }

  /**
   * [run description]
   * @return {Promise} [description]
   */
  run() {
    debug('Starting ActJS');
    /* Variables */
    let wires       = {}
      , middleware  = [];

    /* Loop middlewares and add them to the wires */
    this.wires.forEach(middleware => {
      wires[middleware] = this.container.get(middleware);
    });

    /* @type {Array} */
    middleware = sort(wires)
      .filter(wire => !!wire.middleware)
      .map(wire => this.container.resolver(wire.middleware, true));

    /* @type {Function} */
    const fn = compose(middleware);

    debug('Running %d middleware', middleware.length);
    /* Call the method and return the Promise */
    return fn().then(() => this.container);
  }

  /**
   * [resolve description]
   * @param  {Object}   wire
   * @param  {Array}    definitions
   * @param  {Boolean}  resolve
   * @return {Act}
   */
  define(wire, definitions, resolve) {
    /* Assert */
    if (typeof definitions !== 'object') {
      return this;
    }

    /* Loop the object and create resolvers */
    for (let key in definitions) {
      /* Set the item in our registry */
      this.container.register(`${wire.code}.${key}`, definitions[key], resolve);
    }
    return this;
  }
}

/**
 * [createInstance description]
 * @param  {Mixed} wire       [description]
 * @return {Object|Array}     [description]
 */
function createInstance(wire) {
  /* Create middleware from object or function/class */
  if (Array.isArray(wire)) {
    return wire.map(createInstance);
  } else if (typeof wire === 'object') {
    return Object.create(wire);
  } else if (typeof wire === 'function') {
    return new wire(); //jshint ignore:line
  }
  /* Finally throw the error */
  throw new Error(`Invalid wire type, should be function/object: ${typeof wire}`);
}

/* Expose `Act` */
module.exports = Act;
