
'use strict';

/* Dependencies */
const debug = require('debug')('actjs:boot')
  , escape  = require('escape-regexp')
  , findup  = require('findup-sync')
  , format  = require('util').format
  , path    = require('path')
  , union   = require('lodash.union');

/* Expose `boot` */
module.exports = boot;

/**
 * @method boot
 * @param  {String} fileName
 * @param  {Object} options
 * @return {Array}
 */
function boot(fileName, options) {
  /* @type {Object} */
  options = options || {};
  debug('Booting middleware from: %s', fileName);

  /* Set proper options.types */
  if (typeof options.types === 'undefined') {
    options.types = ['dependencies'];
  } else if (!Array.isArray(options.types)) {
    options.types = [options.types];
  }

  /**
   * Load a single wire, try to determine its folder in the
   * node_modules folder and require the module. Make sure
   * the package json exposes the proper file.
   *
   * @param  {String} moduleName  The name of the wire to load
   * @return {Object}
   */
  return loadMiddlewares(fileName, options).map(moduleName => {
    /* @type {String} The folder name */
    const folderName = findup(path.join('node_modules', moduleName), {
      cwd: options.cwd || process.cwd(),
    });

    /* @type {Object} The middlewarein instance */
    try {
      return require(folderName);
    } catch (e) {
      throw new Error(format('Unable to require middleware `%s`: %s', moduleName, e.message));
    }
  });
}

/**
 * @method loadMiddlewares
 * @param  {String} fileName
 * @param  {Object} options
 * @return {Array}
 */
function loadMiddlewares(fileName, options) {
  /* Variables */
  let config = {}
    , result = [];

  try {
    config = require(fileName);
  } catch (e) {
    debug('Unable to load: %s', fileName);
    throw e;
  }

  /* @type {RegExp} */
  const include = createRegExp(options.include || [])
    , exclude   = createRegExp(options.exclude || ['^actjs$']);

  /* loop dependency types and scan for wires. */
  for (let type of options.types) {
    /* Skip unset configurations */
    if (typeof config[type] !== 'object') {
      continue;
    }
    /* jshint ignore:start */
    /* @type {Array} */
    result = union(result, Object.keys(config[type]).filter(name => {
      return name.match(include) && !name.match(exclude);
    }));
    /* jshint ignore:end */
  }
  return result;
}

/**
 * @method createRegExp
 * @param  {Array|String} regexps An array containing all needed regexps as a string
 * @return {RegExp|null}
 */
function createRegExp(regexps) {
  if (!regexps || !regexps.length) {
    return null;
  }
  /* @type {Array} */
  let result = [];
  for (let regexp of regexps) {
    if (regexp.substr(0, 1) === '^' || regexp.substr(-1) === '$') {
      result.push(regexp);
    } else {
      result.push(escape(regexp));
    }
  }
  /* Return the compiled regexp */
  return new RegExp('(' + result.join('|') + ')');
}
