
'use strict';

/* Dependencies */
const debug = require('debug')('actjs:container')
  , assert  = require('assert')
  , get     = require('lodash.get')
  , set     = require('lodash.set');

/**
 * @class Container
 */
class Container {
  /**
   * The constructor creates an immutable data
   * object on this class. It holds all dependencies
   * and resolved methods.
   * @method  constructor
   */
  constructor() {
    debug('Creating Container');
    /* @type {Object} */
    Object.defineProperty(this, 'data', {value: {}});
  }

  /**
   * [get description]
   * @param  {String} key   The key of the item in the registry
   * @return {Mixed}        The value or resolved value of the item in the registry
   */
  get(key) {
    /* Test validity of string */
    assert(typeof key === 'string', `Key is not a valid string, type ${typeof key} given`);

    /* @type {Mixed} */
    let result = get(this.data, key);

    /* Assert the result is not undefined */
    assert(typeof result !== 'undefined', `${key} not defined`);

    /* Return the result */
    if (typeof result === 'function' && result.name.endsWith('__resolver__')) {
      /* Resolve the result and store it */
      result = result();
      /* Set the data in our registry */
      set(this.data, key, result);
    }
    /* Return the result */
    return result;
  }

  /**
   * Check for existence in the data object.
   * @param  {String}  key  The key of the item to test
   * @return {Boolean}      The result of the evaluation
   */
  has(key) {
    /* Enforce string keys and defined values */
    assert(typeof key === 'string', `Key should be a valid string, ${typeof key} given`);

    /* Return the tested value */
    return typeof get(this.data, key) !== 'undefined';
  }

  /**
   * Register a service or constant within the container, if the definition
   * is an Array it will create a __resolver__ function for it and set
   * this in the registry.
   * @param  {String}  key          The key of the item to register
   * @param  {Mixed}   definition   The actual definition of for the resolver, if array
   * @param  {Boolean} resolve      Whether to resolve the function on runtime
   * @return {Object}               [description]
   */
  register(key, definition, resolve) {
    /* We dont allow overwriting values */
    assert(!this.has(key), `${key} is already defined`);

    /* Create a resolver function if the definition is an array */
    if (Array.isArray(definition)) {
      definition = Container.__resolver__.bind(null, this, definition, resolve);
    }
    /* Set the definition */
    return this.set(key, definition);
  }

  /**
   * Set an item in the container, this only checks whether it exists, and
   * sets the specified value as is.
   * @param {String} key    The key of the item to set
   * @param {Mixed}  value  The value of the item to set
   */
  set(key, value) {
    /* We dont allow overwriting values */
    assert(!this.has(key), `${key} is already defined`);

    /* Enforce string keys and defined values */
    assert(typeof value !== 'undefined', `Value for ${key} should not be undefined`);

    /* Set value on this.data */
    set(this.data, key, value);

    /* Return instance of Container */
    return this;
  }

  /**
   * Simple proxy method to create a resolver for the current container.
   * @param  {Array}    definition
   * @param  {Boolean}  resolve
   * @return {Function}
   */
  resolver(definition, resolve) {
    return Container.__resolver__.bind(null, this, definition, resolve);
  }

  /**
   * Return a resolved definition
   * @param  {Array}    definition [description]
   * @param  {Boolean}  resolve    [description]
   * @return {Mixed}               [description]
   */
  resolve(definition, resolve) {
    return Container.__resolver__.call(null, this, definition, resolve);
  }

  /**
   * @param  {Container}  container   The container instance
   * @param  {Array}      definition  The definition array
   * @param  {Boolean}    resolve     Whether to resolve the result
   * @return {Mixed}
   */
  static __resolver__(container, definition, resolve) {
    /* Assert input */
    assert(container instanceof Container, 'container is not an instance of Container');
    assert(Array.isArray(definition), `No valid definition array specified, ${typeof definition} given`);

    /**
     * If the definition is a function we reset definition
     * to an array, otherwise pop the last element of the
     * definition.
     * @type {Function}
     */
    let fn = definition.pop();
    if (typeof fn === 'string') {
      fn = container.get(fn);
    }

    /* Assert FN should be a function */
    assert(typeof fn === 'function', `Unable to resolve: callee is not a function, ${typeof fn} given`);

    /* Map definitions and add extra arguments */
    definition = definition
      .map(key => container.get(key))
      .concat(Array.from(arguments).splice(3));

    /* Resolve the function */
    if (resolve) {
      return fn.apply(fn, definition);
    }
    return fn.bind(fn, ...definition);
  }
}

/* Expose `Container` */
module.exports = Container;
