
'use strict';

/* Dependencies */
const debug = require('debug')('actjs:sort')
  , assert  = require('assert');

/* Expose `sort` */
module.exports = sort;

/**
 * Sort the plugins by their dependencies. This sorter applies
 * a topological sort algorithm to the specified wires object.
 * It returns the sorted array
 *
 * @param  {Object} edges
 * @return {Array}
 */
function sort(edges) {
  assert(!Array.isArray(edges) && typeof edges === 'object', 'edges should be an object');
  debug('Starting to sort wires');

  const tmp   = {}
    , sorted  = [];

  /**
   * Loop the plugins and visit each plugin to
   * determine the call graphs.
   */
  for (let code in edges) {
    visit(edges[code]);
  }

  /* Return the sorted array */
  debug('Finished sorting the dependency graph');
  return sorted;

  /**
   * Recursively visit every node and determine cyclomatic dependencies,
   * unmet dependencies and marks wires as visited. This function is nested
   * because it relies on the sorted and tmp arrays.
   *
   * @param   {Function} wire The actual wire to visit.
   * @access  private
   */
  function visit(wire) {
    /* Detect cyclic dependency */
    assert(
      tmp[wire.code] !== true,
      `Cyclic dependency found for ${wire.code}`
    );

    /* Return when the plugin already has been visited. */
    if (tmp[wire.code] === false) {
      return;
    }

    /* Mark `plugin` as temporary */
    tmp[wire.code] = true;

    /* Loop dependencies */
    if (Array.isArray(wire.depends)) {
      for (let dependency of wire.depends) {
        /* Check for unmet dependencies */
        assert(typeof edges[dependency] !== 'undefined', `Unmet dependency ${dependency} for ${wire.code}`);

        /* Recursively visit node */
        visit(edges[dependency]);
      }
    }

    /* Unmark the node as temporary */
    tmp[wire.code] = false;

    /* Finally push node to the sorted array */
    sorted.push(wire);

    debug('Sorted %s', wire.code);
  }
};
